unit hiAVICreator;

interface

uses Kol,Windows,vfw,Share,Debug,MMSystem;

type FourCCstr = string[4];

type //   CallBack 
  TCallMMTimer = packed record
    POPEDX: Byte;
    MOVEAX: Byte;
    SelfPtr: Pointer;
    PUSHEAX: Byte;
    PUSHEDX: Byte;
    JMP: Byte;
    JmpOffset: Integer;
  end;

type
  THIAVICreator = class(TDebug)
   private
    AVIFile:PAVIFile;
    AVIStream,CompStream:PAVIStream;
    intFrameCount:Integer;
    whdHandle:HWND;
    scrWidth, scrHeight:integer;
    intQuality:integer;
    intRate:integer;
    intMMtimer:integer;

    BitmapInfo: PBitmapInfoHeader;
    BitmapInfoSize:Integer;
    BitmapBits:Pointer;
    BitmapSize:Integer;
    PixForm:TPixelFormat;

    CallbackMMT: TCallMMTimer;  // CallBack 

   public
    strAviFileName:string;
    strCodec:string[4];
    strTitle:string[63];
    mmt:integer;

    _prop_DC:integer;
    _prop_FileName:string;
    _prop_Rate:integer;
    _prop_Quality:integer;
    _prop_Title:string;
    _prop_Width:integer;
    _prop_Height:integer;
    _prop_Compression:boolean;
    _prop_ShowDialog:boolean;
    _prop_Cursor:boolean;
    _prop_ColorBit:integer;
    _prop_ResizeHandle:boolean;
    _prop_AutoTimer:boolean;
    _prop_Time:integer;

    _data_DC:THI_Event;
    _data_FileName:THI_Event;
    _data_Title:THI_Event;
    _data_Width:THI_Event;
    _data_Height:THI_Event;
    _data_Codec:THI_Event;
    _data_Quality:THI_Event;
    _data_Rate:THI_Event;
    _event_onError:THI_Event;
    _event_onCount:THI_Event;
    _event_onCodecList:THI_Event;
    _event_onPreview:THI_Event;

    constructor Create;
    destructor Destroy; override;
    function Close: Integer;

    procedure _work_doOpenAviFile(var _Data:TData; Index:word);
    procedure _work_doAddFarame(var _Data:TData; Index:word);
    procedure _work_doPause(var _Data:TData; Index:word);
    procedure _work_doCloseAviFile(var _Data:TData; Index:word);
    procedure _work_doOptions(var _Data:TData; Index:word);
    procedure _work_doCodec(var _Data:TData; Index:word);
    procedure Compressorlist;
    procedure SetFormat(var frmBM, outBM : KOL.PBitmap);
    procedure CaptureHandle(var scrBM: KOL.PBitmap);
    procedure ResizeHandle(whd,w,h: integer);
    procedure MMTCallback(TimerID, Msg: Uint; dwUser, dw1, dw2: DWORD); stdcall;

    procedure InternalGetDIBSizes(Bitmap: HBITMAP; var InfoHeaderSize: integer;
                                  var ImageSize: Longint; PixelFormat: TPixelFormat);
    function InternalGetDIB(Bitmap: HBITMAP; Palette: Hpalette;
                            var BitmapInfo; var Bits; PixelFormat: TPixelFormat): boolean;
    procedure InitializeBitmapInfoHeader(Bitmap: HBITMAP; var Info: TBitmapInfoHeader;
                                         PixelFormat: TPixelFormat);

end;

implementation

var AVIInitCount : Integer;

function FourCC(S: ShortString): Integer;
begin
  Move(S[1],Result,4);
end;

{
function FourCCToString(fcc: DWord): string;
var fccptr: ^byte;
    tmp:array[0..3] of char;
begin
  fccptr := @fcc;
  tmp[0] := chr(fccptr^); inc(fccptr);
  tmp[1] := chr(fccptr^); inc(fccptr);
  tmp[2] := chr(fccptr^); inc(fccptr);
  tmp[3] := chr(fccptr^);
  Result := tmp[0] + tmp[1] + tmp[2] + tmp[3];
end;
}

function FourCCToString(fcc: DWord): FourCCstr;
var rtrStr: PChar;
begin
  rtrStr := @fcc;
  Result := rtrStr;
end;

function WideStringToString(const ws: WideString; codePage: Word): AnsiString;
var l: integer;
begin
  if ws = '' then
    Result := ''
else
  begin
    l := WideCharToMultiByte(codePage,
      WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR,
      @ws[1], -1, nil, 0, nil, nil);
    SetLength(Result, l - 1);
    if l > 1 then
      WideCharToMultiByte(codePage,
        WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR,
        @ws[1], -1, @Result[1], l - 1, nil, nil);
  end;
end;

function StringToWideString(const s: String; codePage: Word): WideString;
var len: integer;
begin
  Result := '';
  if s = '' then exit;
  len := MultiByteToWideChar(codePage, MB_PRECOMPOSED, PChar(@s[1]), -1, nil, 0);
  SetLength(Result, len - 1);
  if len <= 1 then exit;
  MultiByteToWideChar(CodePage, MB_PRECOMPOSED, PChar(@s[1]), -1, PWideChar(@Result[1]), len);
end;


function THIAVICreator.InternalGetDIB(Bitmap: HBITMAP; Palette: Hpalette;
         var BitmapInfo; var Bits; PixelFormat: TPixelFormat): boolean;

var OldPal: Hpalette;
    DC: HDC;
begin
  InitializeBitmapInfoHeader(Bitmap, TBitmapInfoHeader(BitmapInfo), PixelFormat);
  OldPal := 0;

  DC := CreateCompatibleDC(0);
  try
    if (Palette <> 0) then
    begin
      OldPal := SelectPalette(DC, Palette, false);
      realizepalette(DC);
    end;
    GDIFlush;
    Result := (GetDIBits(DC, Bitmap, 0, abs(TBitmapInfoHeader(BitmapInfo).biHeight),
      @Bits, TBitmapInfo(BitmapInfo), DIB_RGB_COLORS) <> 0);
  finally
    if (OldPal <> 0) then
      SelectPalette(DC, OldPal, false);
    DeleteDC(DC);
  end;
end;

procedure THIAVICreator.InternalGetDIBSizes(Bitmap: HBITMAP; var InfoHeaderSize: integer;
          var ImageSize: Longint; PixelFormat: TPixelFormat);

var Info: TBitmapInfoHeader;
begin
  InitializeBitmapInfoHeader(Bitmap, Info, PixelFormat);
  // Check for palette device format
  if (Info.biBitCount > 8) then
  begin
    // Header but no palette
    InfoHeaderSize := SizeOf(TBitmapInfoHeader);
    if ((Info.biCompression and BI_BITFIELDS) <> 0) then
      inc(InfoHeaderSize, 12);
  end else
    // Header and palette
    InfoHeaderSize := SizeOf(TBitmapInfoHeader) + SizeOf(TRGBQuad) * (1 shl Info.biBitCount);
  ImageSize := Info.biSizeImage;
end;

procedure THIAVICreator.InitializeBitmapInfoHeader(Bitmap: HBITMAP;
            var Info: TBitmapInfoHeader; PixelFormat: TPixelFormat);
// From graphics.pas, "optimized" for our use
var DIB: TDIBSection; Bytes: Integer;
  function AlignBit(Bits, BitsPerPixel, Alignment: Cardinal): Cardinal;
  begin
    Dec(Alignment);
    Result := ((Bits * BitsPerPixel) + Alignment) and not Alignment;
    Result := Result shr 3;
  end;
begin
  DIB.dsbmih.biSize := 0;
  Bytes := GetObject(Bitmap, SizeOf(DIB), @DIB);
  If (Bytes = 0) then _hi_onEvent(_event_onError,'Error: Invalid bitmap');
  If (Bytes >= (SizeOf(DIB.dsbm) + SizeOf(DIB.dsbmih))) and
  (DIB.dsbmih.biSize >= SizeOf(DIB.dsbmih)) then Info := DIB.dsbmih
  else begin
    FillChar(Info, SizeOf(Info), 0);
    With Info, DIB.dsbm do begin
      biSize := SizeOf(Info);
      biWidth := bmWidth; biHeight := bmHeight;
    end;
  end;

  Case PixelFormat of
    pf1bit: Info.biBitCount := 1;
    pf4bit: Info.biBitCount := 4;
    pf8bit: Info.biBitCount := 8;
    pf15bit: Info.biBitCount := 15;
    pf16bit: Info.biBitCount := 16;
    pf24bit: Info.biBitCount := 24;
    pf32bit: Info.biBitCount := 32;
  else _hi_onEvent(_event_onError,'Error: Invalid pixel foramt');
  end;
    Info.biPlanes := 1;
    Info.biCompression := BI_RGB; // Always return data in RGB format
    Info.biSizeImage := AlignBit(Info.biWidth, Info.biBitCount, 32) *
    Cardinal(Abs(Info.biHeight));
end;

constructor THIAVICreator.Create;
begin
  inherited Create;
  if AVIInitCount=0 then AVIFileInit;
  Inc(AVIInitCount);
  AVIFile:=nil;
  AVIStream:=nil;
  CompStream:=nil;
  mmt := 0;
  //Compressorlist;
  CallbackMMT.POPEDX := $5A;
  CallbackMMT.MOVEAX := $B8;
  CallbackMMT.SelfPtr := Self;
  CallbackMMT.PUSHEAX := $50;
  CallbackMMT.PUSHEDX := $52;
  CallbackMMT.JMP := $E9;
  CallbackMMT.JmpOffset := Integer(@THIAVICreator.MMTCallback) - Integer(@CallbackMMT.JMP) - 5;
end;

destructor THIAVICreator.Destroy;
begin
  TimeKillEvent(mmt);
  Close;
  Dec(AVIInitCount);
  if AVIInitCount=0 then  AVIFileExit;
  inherited Destroy;
end;

procedure THIAVICreator.MMTCallback;
var dt: TData;
begin
  _work_doAddFarame(dt,0);
end;

procedure THIAVICreator.ResizeHandle(whd,w,h: integer);
begin
  if whd <> 0 then
    begin
      SetWindowPos(whd,0,0,0,w,h,SWP_NOMOVE or SWP_NOZORDER);
    end;
end;

procedure THIAVICreator.CaptureHandle(var scrBM: KOL.PBitmap);
var  dc:HDC;
     r:TRect;
     bm: KOL.PBitmap;
begin
   if whdHandle = 0 then begin
      dc := GetDC(0);
      bm := NewBitmap(ScreenWidth,ScreenHeight);
   end else begin
      dc := GetWindowDC(whdHandle);
      GetWindowRect(whdHandle,r);
      bm := NewBitmap(r.Right-r.Left,r.Bottom-r.Top);
   end;
   BitBlt(bm.Canvas.Handle,0,0,bm.Width,bm.Height,DC,0,0,SRCCOPY);
   ReleaseDC(whdHandle,dc);
   scrBM.Assign(bm);
   //  -   
   //scrBM.PixelFormat := PixForm;
   //_hi_onEvent(_event_onPreview,scrBM);
   bm.RemoveCanvas;
   bm.Free;
end;

procedure THIAVICreator.SetFormat(var frmBM, outBM : KOL.PBitmap);
var   ACursor: HICON;
      Pt: TPoint;
      CurInfo: tagCURSORINFO;
      IcoInfo: _ICONINFO;
      t:tagWINDOWPLACEMENT;
      hw,hh: integer;
      tmpBM: KOL.PBitmap;
begin
   if (frmBM = nil) or frmBM.Empty then exit;
   tmpBM := NewBitmap(0,0);
   tmpBM.Assign(frmBM);
   tmpBM.PixelFormat := PixForm;

   //   
   if _prop_Cursor then
   begin
     // hw  hh      handle
     hw := 0; hh := 0;
     case _prop_DC of
     1: begin
         if whdHandle <> 0 then
           begin
             t.length := sizeof(t);
             GetWindowPlacement(whdHandle,@t);
             hw := t.rcNormalPosition.Left; //   
             hh := t.rcNormalPosition.Top;  //   
           end;
        end;
     end;

     CurInfo.cbSize := SizeOf(CurInfo);
     GetCursorInfo(CurInfo);
     ACursor := CurInfo.hCursor;
     Pt := CurInfo.ptScreenPos;
     GetIconInfo (ACursor,IcoInfo);
     //  
     case _prop_DC of
     1:  DrawIcon(tmpBM.Canvas.Handle,(Pt.X-hw) - Integer(IcoInfo.xHotspot),
                 (Pt.Y-hh) - Integer(IcoInfo.yHotspot),ACursor);
     0:  DrawIcon(tmpBM.Canvas.Handle,Pt.X - Integer(IcoInfo.xHotspot),
                  Pt.Y - Integer(IcoInfo.yHotspot),ACursor);
     end;
   end;

   //   
   SetStretchBltMode(outBM.Canvas.Handle, STRETCH_HALFTONE);
   //   
   StretchBlt(outBM.Canvas.Handle,0, 0,scrWidth, scrHeight,
              tmpBM.Canvas.Handle,0, 0,tmpBM.width,
              tmpBM.height, SRCCOPY);
   outBM.RemoveCanvas;
  tmpBM.Free;
end;

procedure THIAVICreator._work_doOpenAviFile;
var
 AVIStreamInfo : TAVIStreamInfo;
 CompOptions   : TAVICompressOptions;
 PCompOptions  : PAVICompressOptions;
 clsidHandler  : PClsID;
 intErr        : integer;
 tmpBM,scrBM   : KOL.PBitmap;
begin
  //  
  scrWidth   := ReadInteger(_Data,_data_Width,_prop_Width);
  scrHeight  := ReadInteger(_Data,_data_Height,_prop_Height);
  //  
  intQuality := ReadInteger(_Data,_data_Quality,_prop_Quality);
  intRate    := ReadInteger(_Data,_data_Rate,_prop_Rate);
  // 
  strCodec   := ReadString(_Data, _data_Codec, 'XVID');
  // ID 
  strTitle   := ReadString(_Data, _data_Title, _prop_Title);
  //    
  //if _prop_AutoTimer then intMMtimer := round (1000 / intRate);//_prop_Time;
  intMMtimer :=  _prop_Time;
  //   
  case _prop_ColorBit of
   0: PixForm := pf1bit;
   1: PixForm := pf4bit;
   2: PixForm := pf8bit;
   3: PixForm := pf15bit;
   4: PixForm := pf16bit;
   5: PixForm := pf24bit;
   6: PixForm := pf32bit;
  end;

  //  
  intFrameCount := 0;
  //   
  strAviFileName := ReadString(_Data, _data_FileName, _prop_FileName);
  //    
  DeleteFile(PChar(strAviFileName));
  //  ,   
   case _prop_DC of
     1: begin
          whdHandle := ReadInteger(_Data,_data_DC,0);
          if _prop_ResizeHandle then ResizeHandle(whdHandle,scrWidth,scrHeight);
          scrBM := NewBitmap(scrWidth,scrHeight);
          tmpBM := NewBitmap(0,0);
          CaptureHandle(tmpBM);
          SetFormat(tmpBM,scrBM);
          //_hi_onEvent(_event_onPreview,scrBM);
        end;
     0: begin
          //  
          tmpBM := NewBitmap(0,0);
          tmpBM.Assign(ToBitmapEvent(_data_DC));
          //  -   
          tmpBM.PixelFormat := PixForm;
          //   
          scrBM := NewBitmap(scrWidth,scrHeight);
          //   
          SetStretchBltMode(scrBM.Canvas.Handle, STRETCH_HALFTONE);
          //   
          StretchBlt(scrBM.Canvas.Handle, 0, 0, scrWidth, scrHeight, tmpBM.Canvas.Handle,
                     0, 0, tmpBM.width, tmpBM.height, SRCCOPY);
          scrBM.RemoveCanvas;
          //_hi_onEvent(_event_onPreview,scrBM);
        end;
   end;


 intErr:=AVIFileOpen(AVIFile,PChar(strAviFileName),OF_CREATE or OF_WRITE,nil);
 if intErr<>0 then
   begin
    Close;
    _hi_onEvent(_event_onError, 'Error open avi file');
    Exit;
   end;

 InternalGetDIBSizes(scrBM.Handle,BitmapInfoSize,BitmapSize,PixForm);
 BitmapInfo := AllocMem(BitmapInfoSize);
 BitmapBits := AllocMem(BitmapSize);
 InternalGetDIB(scrBM.Handle,0,BitmapInfo^,BitmapBits^,PixForm);

 FillChar(AVIStreamInfo,SizeOf(AVIStreamInfo),0);
 with AVIStreamInfo do
 begin
  fccType:=streamtypeVIDEO;
  if _prop_Compression then fccHandler:=FourCC(strCodec)
  else fccHandler:=0;
  dwFlags:=0;
  dwScale:=1; //    = dwRate / dwScale
  dwRate:=intRate; // fps 1 to 30
  dwStart:=0;
  dwLength:=1;
  dwQuality:=intQuality; //0-10000  
  dwSampleSize:=0; // 
  rcFrame.Right := BitmapInfo^.biWidth;
  rcFrame.Bottom := BitmapInfo^.biHeight;
  Move(strTitle[1],szName,63); szName[63]:=#0;
  dwSuggestedBufferSize := BitmapInfo^.biSizeImage;   //  

 end;
 intErr:=AVIFileCreateStream(AVIFile,AVIStream,@AVIStreamInfo);
 if intErr<>0 then begin Close; _hi_onEvent(_event_onError,'Error create file stream'); Exit; end;

 if _prop_Compression then
 begin
  FillChar(CompOptions,SizeOf(CompOptions),0);
  with CompOptions do  //AVICompressOptions
  begin
   fccType:=AVIStreamInfo.fccType;
   fccHandler:=AVIStreamInfo.fccHandler;
   dwKeyFrameEvery:=1; //SpinEditKeyFrames.Value; // .
   dwQuality:=AVIStreamInfo.dwQuality;
   dwBytesPerSecond:=1024 * 100; //  ;
   dwFlags:=AVICOMPRESSF_KEYFRAMES or AVICOMPRESSF_DATARATE;
   dwInterleaveEvery := 1;
   lpFormat:=BitmapInfo;
   cbFormat:=BitmapInfo.biSizeImage;

   if _prop_ShowDialog then
   begin
    PCompOptions:=@CompOptions;
    dwFlags:=dwFlags or AVICOMPRESSF_VALID;
    if not AVISaveOptions(ReadHandle,ICMF_CHOOSE_KEYFRAME or ICMF_CHOOSE_DATARATE,1,AVIStream,PCompOptions) then
    begin
     Close;
     DeleteFile(PChar(strAviFileName));
     _hi_onEvent(_event_onError, 'Error save options');
     Exit;
    end;
   clsidHandler:=nil;
   intErr:=AVIMakeCompressedStream(CompStream,AVIStream,PCompOptions,clsidHandler)
   end else begin
     PCompOptions:=@CompOptions;
     clsidHandler:=nil;
     intErr:=AVIMakeCompressedStream(CompStream,AVIStream,PCompOptions,clsidHandler);
   end;

   if intErr<>0 then begin Close; _hi_onEvent(_event_onError,'Error make compressed stream'); Exit; end;
  end;

 end
 else
 begin
  CompStream:=AVIStream;
  AVIStream:=nil;
 end;
   intErr:=AVIStreamSetFormat(CompStream,0,BitmapInfo,BitmapInfoSize);
   if intErr<>0 then begin Close; _hi_onEvent(_event_onError,'Error create stream set format'); Exit; end;

  //     
  if _prop_AutoTimer then
  mmt := TimeSetEvent(intMMtimer, 0, @CallbackMMT, DWORD(Self),
                      TIME_PERIODIC or TIME_CALLBACK_FUNCTION);

  scrBM.Free; tmpBM.Free;
end;

//     0- 1-
procedure THIAVICreator._work_doPause;
begin
  if _prop_AutoTimer then
    begin
      case ToInteger(_Data) of
      0: begin TimeKillEvent(mmt); mmt := 0; end;
      1: mmt := TimeSetEvent(intMMtimer, 0, @CallbackMMT, DWORD(Self),
                             TIME_PERIODIC or TIME_CALLBACK_FUNCTION);
      end;
    end;
end;

//    
function THIAVICreator.Close;
begin
  Result:=0;
  intFrameCount := 0;
  if _prop_AutoTimer then  begin TimeKillEvent(mmt); mmt := 0; end;
  Sleep(500);
  if CompStream<>nil then begin Result:=Result or AVIStreamRelease(CompStream); CompStream:=nil; end;
  if AVIStream<>nil  then begin Result:=Result or AVIStreamRelease(AVIStream);  AVIStream:=nil;  end;
  if AVIFile<>nil    then begin Result:=Result or AVIFileRelease(AVIFile);      AVIFile:=nil;    end;
  intFrameCount := 0;
end;

//    
procedure THIAVICreator._work_doAddFarame;
var tmpBM,scrBM : KOL.PBitmap;
begin
  //  
  case _prop_DC of
     1: begin
          whdHandle := ReadInteger(_Data,_data_DC,0);
          if _prop_ResizeHandle then ResizeHandle(whdHandle,scrWidth,scrHeight);
          scrBM := NewBitmap(scrWidth,scrHeight);
          tmpBM := NewBitmap(0,0);
          CaptureHandle(tmpBM);
          SetFormat(tmpBM,scrBM);
        end;
    0: begin
         //  
         tmpBM := NewBitmap(0,0);
         tmpBM.Assign(ToBitmapEvent(_data_DC));
         //  -   
         tmpBM.PixelFormat := PixForm;
         scrBM := NewBitmap(scrWidth,scrHeight);
         SetFormat(tmpBM,scrBM); // 
       end;
  end;

  //  
  InternalGetDIB(scrBM.Handle,0,BitmapInfo^,BitmapBits^,PixForm);
  if AVIStreamWrite(CompStream,intFrameCount,1,BitmapBits,BitmapSize,
                  AVIIF_KEYFRAME  {0} ,nil,nil)<>AVIERR_OK then     // AVIIF_KEYFRAME
  begin
    _hi_onEvent(_event_onError,'Could not add frame');
    exit;
  end;

  _hi_onEvent(_event_onCount,intFrameCount);
  Inc(intFrameCount);
  _hi_onEvent(_event_onPreview,scrBM);

  tmpBM.Free; scrBM.Free;
end;

procedure THIAVICreator._work_doCloseAviFile;
begin
  Close;
  //tmpBitmap.Free; scrBitmap.Free;
end;

procedure THIAVICreator._work_doOptions;
begin

end;

procedure THIAVICreator._work_doCodec;
begin
  Compressorlist;
end;



procedure THIAVICreator.Compressorlist;
var ii: TICINFO;
    i: DWord;
    ic: THandle;
    BitmapInfoHeader: TBitmapInfoHeader;
    Name: WideString;
    j: integer;
begin
  FillChar(BitmapInfoHeader, SizeOf(BitmapInfoHeader), 0);
  with BitmapInfoHeader do
  begin
    biSize := SizeOf(BitmapInfoHeader);
    biWidth := _prop_Width; //fWidth;
    biHeight := _prop_Height; //fHeight;
    biPlanes := 1;
    biCompression := BI_RGB;
    case _prop_ColorBit of
     0: biBitCount := 1;
     1: biBitCount := 4;
     2: biBitCount := 8;
     3: biBitCount := 15;
     4: biBitCount := 16;
     5: biBitCount := 24;
     6: biBitCount := 32;
    end;
    //biBitCount := 24; //Word(pf24Bit);//BitCounts[fPixelFormat];
  end;

  ii.dwSize := SizeOf(ii);
  for i := 0 to 200 do //what's a safe number to get all?
  begin
    if ICInfo(ICTYPE_VIDEO, i, @ii) then
    begin
      ic := ICOpen(ICTYPE_VIDEO, ii.fccHandler, ICMODE_QUERY);
      try
        if ic <> 0 then
        begin
          if ICCompressQuery(ic, @BitmapInfoHeader, nil) = 0 then
          begin
            ICGetInfo(ic, @ii, SizeOf(ii));
            //can the following be done any simpler?
            Name := '';
            for j := 0 to 15 do
              Name := Name + ii.szName[j];
              _hi_onEvent (_event_onCodecList, FourCCToString(ii.fccHandler) +
                           ' ' + WideStringToString(Name,1251));
          end;
        end;
      finally
        ICClose(ic);
      end;
    end;
  end;
end;


end.